/***********************************************************************
 * File: unit_test.c
 * 	(Extracted from
 *	   "$Id: unit_test.c,v 1.5 2009/05/30 14:05:32 maoyam Exp maoyam $")
 *
 * Abstracts:
 *  This C source implements a test program for "EUC_code_counter.c".
 *
 * Copyright (C) 2005-2007 Maoyam Tokyo, Japan (mailto:maoyam@mail.goo.ne.jp)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 **********************************************************************/


/* include the library definitions */
#include "EUC_code_counter.h"

/* Test program for "count_euc_code_characters()"
 * and "substr_of_euc_code_string()" static functions
 */
#define EXIT_NORMAL						0
#define EXIT_USAGE_ERROR				1
#define EXIT_FAILED_OPEN_FILE			2
#define EXIT_FILE_IS_NOT_EUC_ENCODED	3

int main(int ac, char **av)
{
	size_t	buf_len, gprof_cnt, gprof_iter_max;
	FILE	*fp = stdin;
	char	buf[GET_EUC_JP_STR_BUF_SZ(2048)];

	for (gprof_iter_max = (size_t)((ac == 3 && strcmp(av[2], "-P") == 0)?
																3300: 1
									)
		, gprof_cnt = 0
	; gprof_cnt < gprof_iter_max
	; ++gprof_cnt
	) {	/* iterate the unit test for gprof profiling... */

	/* count up ASCII control codes... */
	{	EUC_code_character_class_counts counts;
		size_t	i;

		memset(buf, (int)'\0', sizeof(buf)); /* initialize buffer */

		for (i = 0; i < 0x20; ++i) {
			buf[i]	= ((i == 0x1f)?	ASCII_CNTL_DEL : (char)i + 1);
		}

		printf("buf[ (len = %lu) ] = { ", (unsigned long)strlen(buf));
		for (i = 0, buf_len = strlen(buf); i < buf_len; ++i) {
			printf("%s%x", ((i == 0)? " ": ", "), (unsigned)(buf[i]));
		}
		printf(" }\n"); (void)fflush(stdout);

		i = count_euc_code_characters(
							  (unsigned char *)buf
							, buf_len
							, (EUC_code_character_class_counts *)
									memset(&counts, '\0', sizeof(counts))
							);

		printf(	"count_euc_code_characters(buf@%p "
				", buf_len = %ld, &counts@%p) returns"
				" %ld. counts are {"
				" count_of_ASCII = %ld "
				", count_of_JIS_X_0208_1983 = %ld "
				", count_of_JIS_X_0201_1976 = %ld "
				", count_of_JIS_X_0212_1990 = %ld "
				", count_of_UNKNOWN = %ld "
				", count_of_new_lines = %ld "
				", count_of_controls = %ld "
				", length_of_bytes = %ld "
				"}\n"
				"\n"
				,			buf
				, (long)buf_len, (void *)&counts
				, (long)i
				, (long)counts.count_of_ASCII
				, (long)counts.count_of_JIS_X_0208_1983
				, (long)counts.count_of_JIS_X_0201_1976
				, (long)counts.count_of_JIS_X_0212_1990
				, (long)counts.count_of_UNKNOWN
				, (long)counts.count_of_new_lines
				, (long)counts.count_of_controls
				, (long)counts.length_of_bytes
				); (void)fflush(stdout)
				;
	}

	/* parse command options... */
	if (ac == 1) {	fp	= stdin; /* input from standard input stream */	}
	else if (ac == 2 || ac == 3) {
		/* input from the specified file */
		if ((fp = fopen(av[1], "r")) == NULL) {
			/* FAILED to open the file (read only mode) */
			fprintf(stderr
				, "FATAL: failed to open \"%s\".\n", av[1]
				); (void)fflush(stderr)
				;
			exit(EXIT_FAILED_OPEN_FILE); return EXIT_FAILED_OPEN_FILE;
		}
	}
	else {	/* FAILED: usage error */
		fprintf(stderr, "Usage %s <EUC code file>\n", av[0])
		; (void)fflush(stderr)
		;
		exit(EXIT_USAGE_ERROR); return EXIT_USAGE_ERROR;
	}

	/* process the each lines in input stream... */
	while (fgets((char *)memset(buf, (int)'\0', sizeof(buf))
				, (int)sizeof(buf)
				, fp
				) != NULL
	) {
		EUC_code_character_class_counts	counts;
		size_t							total_charas, n, m = 1
										, substr_sz, width
										;

		/* skip the blank lines, and eliminate the new-lines */
		buf_len			= strlen(buf);
		while ((buf_len > 0)
		&& (buf[buf_len - 1] == '\n' || buf[buf_len - 1] == '\r')
		) {					buf_len--;	}
		if (buf_len == 0) {	continue;	}
		buf[buf_len]	= '\0';

		/* Try to "count_euc_code_characters()" function... */
		total_charas = count_euc_code_characters(
									  (unsigned char *)buf
									, strlen(buf)
									, (EUC_code_character_class_counts *)
											memset(&counts, (int)'\0'
													, sizeof(counts)
													)
									);
		printf(	"count_euc_code_characters(\"%s\" "
				"(buf_len = %ld), %ld, %p) returns"
				" %ld. counts are {"
				" count_of_ASCII = %ld "
				", count_of_JIS_X_0208_1983 = %ld "
				", count_of_JIS_X_0201_1976 = %ld "
				", count_of_JIS_X_0212_1990 = %ld "
				", count_of_UNKNOWN = %ld "
				", count_of_new_lines = %ld "
				", count_of_controls = %ld "
				", length_of_bytes = %ld "
				"}\n"
				,			buf
				, (long)buf_len, (long)strlen(buf)
				, (void *)&counts
				, (long)total_charas
				, (long)counts.count_of_ASCII
				, (long)counts.count_of_JIS_X_0208_1983
				, (long)counts.count_of_JIS_X_0201_1976
				, (long)counts.count_of_JIS_X_0212_1990
				, (long)counts.count_of_UNKNOWN
				, (long)counts.count_of_new_lines
				, (long)counts.count_of_controls
				, (long)counts.length_of_bytes
				); (void)fflush(stdout)
				;
		if (counts.count_of_UNKNOWN > 0) {
			/* FAILED: input stream isn't EUC-JP encoding */
			fprintf(stderr
				,	"FAILED: \"%s\" file isn't "
					"encoded in EUC-code (for Japanese).\n"
				,	(fp == stdin)? "(stdin)": av[1]
				); (void)fflush(stderr)
				;
			exit(EXIT_FILE_IS_NOT_EUC_ENCODED)
			; return EXIT_FILE_IS_NOT_EUC_ENCODED
			;
		}
		/* try to count characters of "EUC to SJIS" converted... */
		printf(	"The number of \"EUC to SJIS\" converted characters is "
				"%ld.\n"
				, (long)count_euc_characters_as_sjis((unsigned char *)buf
													, strlen(buf)
													, &counts
													)
				)
		; printf("\n"); (void)fflush(stdout)
		;

		/* Try to "substr_of_euc_code_string()" function... */
		for (n = 0; n <= total_charas; ++n) {
			for (m = 1; m <= total_charas + 1 - n; ++m) {
				char	*substrp, *delimits = "\n", *NULLp = NULL;

				substr_sz = 0; width = 0;
				if (m == 1) {
					/* TEST: get display width */
					substrp
					= substr_of_euc_code_string(
						  (unsigned char *)buf
						, strlen(buf)
						, n
						, (size_t)0
						, (size_t)3
						, delimits
						, (size_t (*)(unsigned char*, size_t, void*))NULLp
						, NULLp
						, &substr_sz
						, &width
						);
					if (substrp == NULL) {
						printf(	"ERROR: substr_of_euc_code_string(%p "
								"\"%s\","
								" %ld, %ld, %ld,"
								" %ld, %p, %p, %p,"
								" %p, %p) "
								"return %p "
								"(substr_sz = %ld, "
								"errno = %d).\n"
								, buf
								, buf
								, (long)strlen(buf)
								, (long)n, 0L
								, 3L, delimits, NULLp
								, NULLp
								, (void *)&substr_sz
								, (void *)&width
								, substrp
								, (long)substr_sz
								, errno
								); (void)fflush(stdout)
								;
					}
					else {
						printf(	"SUCCESS: substr_of_euc_code_string("
								"%p \"%s\","
								" %ld, %ld, %ld,"
								" %ld, %p, %p, %p,"
								" %p, %p) "
								"return %p "
								"(substr_sz = %lu, "
								"width = %lu).\n"
								"\tsub-string is "
								"\"%.*s\".\n"
								, buf, buf
								, (long)strlen(buf)
								, (long)n, 0L
								, 3L, delimits, NULLp
								, NULLp
								, (void *)&substr_sz
								, (void *)&width
								, substrp
								, (unsigned long)substr_sz
								, (unsigned long)width
								, (int)substr_sz
								, substrp
								); (void)fflush(stdout)
								;
					}
				}
				
				/* TEST: get sub-strings */
				substrp
				= substr_of_euc_code_string(
						  (unsigned char *)buf
						, strlen(buf)
						, n
						, m
						, 0
						, delimits
						, (size_t (*)(unsigned char*, size_t, void*))NULLp
						, NULLp
						, &substr_sz
						, &width
						);
				if (substrp == NULL) {
					printf(	"ERROR: "
							"substr_of_euc_code_string(%p "
							"\"%s\","
							" %ld, %ld, %ld,"
							" %ld, %p, %p, %p,"
							" %p, %p) "
							"return %p "
							"(substr_sz = %ld, "
							"errno = %d).\n"
							,			buf
							, buf
							, (long)strlen(buf), (long)n
							, (long)m
							, 0L, delimits, NULLp, NULLp
							, (void *)&substr_sz
							, (void *)&width
							, substrp
							, (long)substr_sz
							, errno
							); (void)fflush(stdout)
							;
				}
				else {
					printf(	"SUCCESS: "
							"substr_of_euc_code_string("
							"%p \"%s\","
							" %lu, %lu, %lu,"
							" %lu, %p, %p, %p,"
							" %p, %p) "
							"return %p "
							"(substr_sz = %lu, "
							"width = %lu).\n"
							"\tsub-string is \"%.*s\".\n"
							, buf, buf
							, (unsigned long)strlen(buf)
							, (unsigned long)n
							, (unsigned long)m
							, 0L
							, delimits, NULLp, NULLp
							, (void *)&substr_sz
							, (void *)&width
							, substrp
							, (unsigned long)substr_sz
							, (unsigned long)width
							, (int)substr_sz, substrp
							); (void)fflush(stdout)
							;
				}
				/* TEST: get sub-strings */
				substrp
				= substr_of_euc_code_string(
						  (unsigned char *)buf
						, strlen(buf)
						, n
						, 0
						, m
						, delimits
						, (size_t (*)(unsigned char*, size_t, void*))NULLp
						, NULLp
						, &substr_sz
						, &width
						);
				if (substrp == NULL) {
					printf(	"ERROR: "
							"substr_of_euc_code_string(%p "
							"\"%s\","
							" %ld, %ld, %ld,"
							" %ld, %p, %p, %p,"
							" %p, %p) "
							"return %p "
							"(substr_sz = %ld, "
							"errno = %d).\n"
							,			buf
							, buf
							, (long)strlen(buf), (long)n
							, 0L
							, (long)m, delimits, NULLp
							,			NULLp
							, (void *)&substr_sz
							, (void *)&width
							, substrp
							, (long)substr_sz
							, errno
							); (void)fflush(stdout)
							;
				}
				else {
					printf(	"SUCCESS: "
							"substr_of_euc_code_string("
							"%p \"%s\","
							" %lu, %lu, %lu,"
							" %lu, %p, %p, %p,"
							" %p, %p) "
							"return %p "
							"(substr_sz = %lu, "
							"width = %lu).\n"
							"\tsub-string is \"%.*s\".\n"
							, buf, buf
							, (unsigned long)strlen(buf)
							, (unsigned long)n
							, 0L
							, (unsigned long)m
							, delimits, NULLp, NULLp
							, (void *)&substr_sz
							, (void *)&width
							, substrp
							, (unsigned long)substr_sz
							, (unsigned long)width
							, (int)substr_sz, substrp
							); (void)fflush(stdout)
							;
				}
			}
			printf("\n"); (void)fflush(stdout);
		}
	}

	/* close the input file */
	if (fp != stdin) {	(void)fclose(fp);	}

	{	
		Trim_EUCJP_spaces_direction_t direct;
		unsigned char				EUCJP_space[] = { 0xa1, 0xa1, '\0' };
		size_t						ax, bx, cx;
		char						*a[]	= { "", " ", NULL, "a" }
									, *b[]	= { "", " ", NULL, "b" }
									, *c[]	= { "", " ", NULL, "c" }
									, str[7], *result, *error_str
									;
			
		printf("\n{{{ test of trim_EUCJP_spaces() start...\n")
		; (void)fflush(stdout)
		;

		for (ax = 0; ax < sizeof(a) / sizeof(a[0]); ++ax) {
			a[2] = b[2] = c[2]	= (char *)EUCJP_space;
		}
		for (ax = 0; ax < sizeof(a) / sizeof(a[0]); ++ax) {
		for (bx = 0; bx < sizeof(b) / sizeof(b[0]); ++bx) {
		for (cx = 0; cx < sizeof(c) / sizeof(c[0]); ++cx) {
		for (direct = (Trim_EUCJP_spaces_direction_t)0
		; direct <= (Trim_EUCJP_spaces_direction_t)0x08
		; direct = (Trim_EUCJP_spaces_direction_t)(direct + 1)
		) {
			(void)snprintf(str, sizeof(str)
							, "%s%s%s"
							, a[ax], b[bx], c[cx]
							);
			errno		= 0;
			printf(	"trim_EUCJP_spaces(direct = 0x%x, "
					"str@%p [%s])\n"
					,			direct
					,	str, str
					); (void)fflush(stdout)
					;
			result		= trim_EUCJP_spaces(direct, str);
			error_str	= strerror(errno);
			printf(	"...returns %p (outputs str@%p [%s]. errno = %d [%s]\n"
					,	result,		str, str
					,	errno, error_str
					); (void)fflush(stdout)
					;
		}}}}

		printf(	"}}} test of trim_EUCJP_spaces() finished!\n")
		; (void)fflush(stdout)
		;
	}

	printf("\n"); (void)fflush(stdout);

	{	char		dst[11];
		size_t		a, b, c, dst_sz;
		char		src[10]
					, *codes[] =	{ ""
									, "a"				/* ASCII */
									, "\x8e\xb2"		/* KANA "i" */
									, "\xa4\xa6"		/* KANJI "u" */
									, "\x8f\xa1\xa1"	/* GAIJI "??" */
									, "\n"				/* ASCII CNTL */
									};

		printf(	"{{{ test of strcpy_EUCJP() starts...\n")
		; (void)fflush(stdout)
		;

		for (a = 0; a < (sizeof(codes) / sizeof(codes[0])); ++a) {
		for (b = (size_t)((a > 0)? 1: 0)
		; b < (sizeof(codes) / sizeof(codes[0]))
		; ++b
		) {
		for (c = (size_t)((b > 0)? 1: 0)
		; c < (sizeof(codes) / sizeof(codes[0]))
		; ++c
		) {
			(void)snprintf(src, sizeof(src)
							, "%s%s%s"
							, codes[a], codes[b], codes[c]
							);
			for (dst_sz = strlen(src) + 1; dst_sz > 0; --dst_sz) {
				size_t	result_len;
				char	*result;

				memset(dst, '@', sizeof(dst) - 1);
				dst[sizeof(dst) - 1]	= '\0';
				printf(	"strcpy_EUCJP(src@%p [%s] "
						"(len=%lu), "
						"dst@%p (len=%lu), "
						"dst_sz=%lu)\n"
						, src, src
						, (unsigned long)strlen(src)
						, dst
						, (unsigned long)strlen(dst)
						, (unsigned long)dst_sz
						); (void)fflush(stdout)
						;
				result	= strcpy_EUCJP((unsigned char *)src
									, (char *)memset(dst, '\0', sizeof(dst))
									, dst_sz
									);
				result_len	= (result == NULL)?	(size_t)ULONG_MAX
								:				strlen(result)
								;
				printf(	"...returns @%p [%s] (len=%lu) (errno=%d)\n"
						, result
						, ((result == NULL)? "(N/A)" : result)
						, (unsigned long)result_len
						, errno
						);
				if ((result != NULL)
				&& memcmp(src, result, result_len) != 0
				) {
					printf("!!!FATAL!!! failed to transfer src to dst!\n");
				}
			}
			printf(	"\n"); (void)fflush(stdout);
		}}}

		printf(	"}}} test of strcpy_EUCJP() finished!\n")
		; (void)fflush(stdout)
		;
	}
	} /* iterate the unit test for gprof profiling... */

	exit(EXIT_NORMAL); return EXIT_NORMAL; /* All OK! */
}
