/***********************************************************************
 * File: abacus_test.c
 * 	(Extracted from
 *	   "$Id: abacus_test.c,v 1.5 2009/06/05 19:15:45 maoyam Exp maoyam $")
 *
 * Abstracts:
 *  This C source file implements test program for `abacus()' API.
 *
 * Copyright (C) 2006-2009 Maoyam Tokyo, Japan (mailto:maoyam@mail.goo.ne.jp)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 **********************************************************************/

#include "abacus.h"
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>

/* in ISO-C99, several functions aren't declared */
extern char *strsignal(int sig);


static void print_commands(size_t cmds_n, Command_t *cmds)
{
	size_t	i, j;

	for (i = 0; i < cmds_n; ++i) {
		printf("commands[%ld]={ opcode=%d, operands[]={"
				, (long)i, cmds[i].opcode
				);
		for (j = 0
		; j < sizeof(cmds->operands) / sizeof(cmds->operands[0])
		; ++j
		) {
			printf("%s @%p", ((j > 0)? ",": ""), cmds[i].operands[j]);
		}
		printf(" }}\n"); (void)fflush(stdout);
	}

	return;
}


int main(int /*@unused@*/ ac, char /*@unused@*/ **av)
{
	char	result[SZ_Numeral_str];
	int		retcode;
	size_t	gprof_cnt, gprof_iter_max;

	for (gprof_iter_max = (size_t)((ac == 2 && strcmp(av[1], "-P") == 0)?
																10000: 1
									)
		, gprof_cnt = 0
	; gprof_cnt < gprof_iter_max
	; ++gprof_cnt
	) {	/* iterate the unit test for gprof prfiling... */

	/* test `Nop_op', `Jmp_op', `Cmp_op', `Set_op' and `Ret_op' mnemonics */
	{
		char	val_in_progress[41][SZ_Numeral_str];
		Command_t commands[] = {
		  /* # 0 */ Nop
		, /* # 1 */ Set(Value(val_in_progress[0]), Register0)
		, /* # 2 */ Set(Value(val_in_progress[1]), Memory(0))
		, /* # 3 */ Set(Value(val_in_progress[2]), Memory(1))
		, /* # 4 */ Set(Register0, Value(
						"-123456789012345678.123456789012345678"
						))
		, /* # 5 */ Set(Memory(0), Value(
						"234567890123456781.234567890123456781"
						))
		, /* # 6 */ Set(Memory(1), Value(
						"499999999999999999.567890123456781234"
						))
		, /* # 7 */ Set(Value(val_in_progress[3]), Register0)
		, /* # 8 */ Set(Value(val_in_progress[4]), Memory(0))
		, /* # 9 */ Set(Value(val_in_progress[5]), Memory(1))
		, /* #10 */ Set(Memory(1), Memory(0)		)
		, /* #11 */ Set(Memory(0), Register0		)
		, /* #12 */ Set(Register0, Value(val_in_progress[5]))
		, /* #13 */ Set(Value(val_in_progress[6]), Register0)
		, /* #14 */ Set(Value(val_in_progress[7]), Memory(0))
		, /* #15 */ Set(Value(val_in_progress[8]), Memory(1))

		#define Series_of_Cmp	19
		#define Calc_commands	29
		, /* #16 */ Jmp(Series_of_Cmp - 16 /* go to `CMP'-s */)
		#define No_error	17
		, /* #17 */ Jmp(Calc_commands - 17 /* go to last */)
		#define Fatal_error	18
		, /* #18 */ Ret(999	/* !!FATAL ERROR!! */	)
		, /* #19 */ Cmp(Register0, Register0
						, Fatal_error - 19,  1, Fatal_error - 19
						)
		, /* #20 */ Cmp(Memory(0), Memory(0)
						, Fatal_error - 20,  1, Fatal_error - 20
						)
		, /* #21 */ Cmp(Memory(1), Memory(1)
						, Fatal_error - 21,  1, Fatal_error - 21
						)
		, /* #22 */ Cmp(Register0, Value(val_in_progress[6])
						, Fatal_error - 22,  1, Fatal_error - 22
						)
		, /* #23 */ Cmp(Memory(0), Value(val_in_progress[7])
						, Fatal_error - 23,  1, Fatal_error - 23
						)
		, /* #24 */ Cmp(Memory(1), Value(val_in_progress[8])
						, Fatal_error - 24,  1, Fatal_error - 24
						)
		, /* #25 */ Cmp(Value(val_in_progress[0])
						, Value(val_in_progress[1])
						, Fatal_error - 25,  1, Fatal_error - 25
						)
		, /* #26 */ Cmp(Register0, Memory(0)
						, Fatal_error - 26, Fatal_error - 26,  1
						)
		, /* #27 */ Cmp(Memory(0), Memory(1)
						, 1, Fatal_error - 27, Fatal_error - 27
						)
		, /* #28 */ Jmp(No_error - 28 /* go back no-error */)

		, /* #29 */ Sub(Value(val_in_progress[5]), Memory(0))
		, /* #30 */ Set(Value(val_in_progress[9]), Register0)
		, /* #31 */ Add(Value(val_in_progress[5]), Memory(0))
		, /* #32 */ Set(Value(val_in_progress[10]), Register0)
		, /* #33 */ Sub(Value(val_in_progress[5]), "0.0")
		, /* #34 */ Set(Value(val_in_progress[11]), Register0)
		, /* #35 */ Add(Value(val_in_progress[5]), "0.0")
		, /* #36 */ Set(Value(val_in_progress[12]), Register0)
		, /* #37 */ Sub(Value(val_in_progress[5]), Memory(1))
		, /* #38 */ Set(Value(val_in_progress[13]), Register0)
		, /* #39 */ Add(Value(val_in_progress[5]), Memory(1))
		, /* #40 */ Set(Value(val_in_progress[14]), Register0)
		, /* #41 */ Sub(Value(val_in_progress[5])
						, Value(val_in_progress[6])
						)
		, /* #42 */ Set(Value(val_in_progress[15]), Register0)
		, /* #43 */ Add(Value(val_in_progress[5])
						, Value(val_in_progress[6])
						)
		, /* #44 */ Set(Value(val_in_progress[16]), Register0)
		, /* #45 */ Sub(Value(val_in_progress[4])
						, Value(val_in_progress[3])
						)
		, /* #46 */ Set(Value(val_in_progress[17]), Register0)
		, /* #47 */ Add(Value(val_in_progress[4])
						, Value(val_in_progress[3])
						)
		, /* #48 */ Set(Value(val_in_progress[18]), Register0)
		, /* #49 */ Sub(Value(val_in_progress[4])
						, Value(val_in_progress[0])
						)
		, /* #50 */ Set(Value(val_in_progress[19]), Register0)
		, /* #51 */ Add(Value(val_in_progress[4])
						, Value(val_in_progress[0])
						)
		, /* #52 */ Set(Value(val_in_progress[20]), Register0)
		, /* #53 */ Sub(Value(val_in_progress[4])
						, Value(val_in_progress[4])
						)
		, /* #54 */ Set(Value(val_in_progress[21]), Register0)
		, /* #55 */ Add(Value(val_in_progress[4])
						, Value(val_in_progress[4])
						)
		, /* #56 */ Set(Value(val_in_progress[22]), Register0)
		, /* #57 */ Sub(Value(val_in_progress[4])
						, Value(val_in_progress[5])
						)
		, /* #58 */ Set(Value(val_in_progress[23]), Register0)
		, /* #59 */ Add(Value(val_in_progress[4])
						, Value(val_in_progress[5])
						)
		, /* #60 */ Set(Value(val_in_progress[24]), Register0)
		, /* #61 */ Sub(Memory(0), Memory(0)		)
		, /* #62 */ Set(Value(val_in_progress[25]), Register0)
		, /* #63 */ Add(Memory(0), Memory(0)		)
		, /* #64 */ Set(Value(val_in_progress[26]), Register0)
		, /* #65 */ Sub(Memory(0), Memory(1)		)
		, /* #66 */ Set(Value(val_in_progress[27]), Register0)
		, /* #67 */ Add(Memory(0), Memory(1)		)
		, /* #68 */ Set(Value(val_in_progress[28]), Register0)
		, /* #69 */ Sub(Memory(0), Value(val_in_progress[0]))
		, /* #70 */ Set(Value(val_in_progress[29]), Register0)
		, /* #71 */ Add(Memory(0), Value(val_in_progress[0]))
		, /* #72 */ Set(Value(val_in_progress[30]), Register0)
		, /* #73 */ Sub(Memory(0), Value(val_in_progress[6]))
		, /* #74 */ Set(Value(val_in_progress[31]), Register0)
		, /* #75 */ Add(Memory(0), Value(val_in_progress[0]))
		, /* #76 */ Set(Value(val_in_progress[32]), Register0)
		, /* #77 */ Sub("-0.0", Value(val_in_progress[3]))
		, /* #78 */ Set(Value(val_in_progress[33]), Register0)
		, /* #79 */ Add("-0.0", Value(val_in_progress[3]))
		, /* #80 */ Set(Value(val_in_progress[34]), Register0)
		, /* #81 */ Sub("0.0", "-0.0"			)
		, /* #82 */ Set(Value(val_in_progress[35]), Register0)
		, /* #83 */ Add("0.0", "-0.0"			)
		, /* #84 */ Set(Value(val_in_progress[36]), Register0)
		, /* #85 */ Sub("0.0", Value(val_in_progress[4]))
		, /* #86 */ Set(Value(val_in_progress[37]), Register0)
		, /* #87 */ Add("0.0", Value(val_in_progress[4]))
		, /* #88 */ Set(Value(val_in_progress[38]), Register0)
		, /* #89 */ Sub("0.0", Value(val_in_progress[5]))
		, /* #90 */ Set(Value(val_in_progress[39]), Register0)
		, /* #91 */ Add("0.0", Value(val_in_progress[5]))
		, /* #92 */ Set(Value(val_in_progress[40]), Register0)

		, /* #93 */ Ret(NORMAL_END			)
		};
		size_t	result_sz = sizeof(result)
				, int_digits = MAX_Digits
				, fra_digits = MAX_Digits
				, memories_n = 2
				, commands_n = 0
				, pc = 0
				, i
				;

		memset(val_in_progress, (int)'\0', sizeof(val_in_progress));
		commands_n = (sizeof(commands) / sizeof(commands[0]));
		printf(	"{{{ Test for `Nop_op', `Jmp_op', `Cmp_op', "
				"`Set_op' and `Ret_op' mnemonics starts...\n"
				); (void)fflush(stdout);

		/* print contents of commands[] */
		print_commands(commands_n, commands)
		; (void)puts(""); (void)fflush(stdout)
		;

		/* initialize buffers... */
		for (i = 0, memset(result, (int)'@', (sizeof(result) - 1))
			, result[sizeof(result) - 1] = '\0'
		; i < (sizeof(val_in_progress) / sizeof(val_in_progress[0]))
		; ++i
		) {	strcpy(val_in_progress[i], result);	}

		/* invoke abacus() API */
		errno	= 0;
		retcode	= abacus(result, result_sz
						, int_digits, fra_digits
						, memories_n, commands_n, commands, &pc
						);
		printf(	"abacus(result@%p, result_sz=%ld, "
				"int_digits=%ld, fra_digits=%ld, "
				"memories_n=%ld, commands_n=%ld, "
				"commands@%p, &pc@%p) "
				"returns %d.\n"
				"(outputs result=[%s](len=%ld), pc=%ld, "
				"errno=%d)"
				,	result, (long)result_sz
				,	(long)int_digits, (long)fra_digits
				,	(long)memories_n, (long)commands_n
				,	(void *)commands, (void *)(&pc)
				,	retcode
				,	result, (long)strlen(result), (long)pc
				,	errno
				);
		if (retcode != NORMAL_END) {
			printf("!!!ERROR!!![%s]", strerror(errno))
			; (void)fflush(stdout)
			;
		}
		(void)puts(""); (void)fflush(stdout);

		/* print values in progress... */
		for (i = 0
		; i < sizeof(val_in_progress) / sizeof(val_in_progress[0])
		; ++i
		) {
			printf(	"val_in_progress[%ld]=[%s](len=%ld)\n"
					,	(long)i, val_in_progress[i]
					,	(long)strlen(val_in_progress[i])
					); (void)fflush(stdout)
					;
		}

		printf(	"}}} Test for `Nop_op', `Jmp_op', `Cmp_op', "
				"`Set_op' and `Ret_op' mnemonics finished.\n"
				); (void)fflush(stdout)
				;
	}
	/* test `Nop_op', `Jmp_op', `Cmp_op', `Set_op' and `Ret_op' mnemonics */
	
	(void)puts(""); (void)fflush(stdout);

	{ /* test `MUL_op' and `Div_op' mnemonics */
		char	*strs[] =	{ "0.0"
							, "0.0123456789"
							, "0.012345678901234567"
							, "-1"
							, "2"
							, "-8"
							, "7"
							, "1234567890"
							, "1234567890.0123456789"
							, "123456789012345678"
							, "123456789012345678.012345678901234567"
							};
		char	vals[5][SZ_Numeral_str], bc_cmd[1024];
		#define	Make_bc_cmd ((snprintf(bc_cmd, sizeof(bc_cmd)			\
										, "printf \"scale=18\n"			\
										 "%s * %s\n"					\
										 "%s / %s\n"					\
										 "%s - ((%s / %s) * %s)\n"		\
										 "m=%s - ((%s / %s) * %s)\n"	\
										 "%s\n"							\
										 "(%s / %s) * %s + m\n"			\
										 "\" | bc -s"					\
										, strs[x], strs[y]				\
										, strs[x], strs[y]				\
										, strs[x], strs[x], strs[y]		\
										,			strs[y]				\
										, strs[x], strs[x], strs[y]		\
										,			strs[y]				\
										, strs[y]						\
										, strs[x], strs[y], strs[y]		\
										)), bc_cmd						\
							)
		Command_t commands[] =	{
					  /* # 0 */ Mul(Value(strs[0]), Value(strs[1]))
					, /* # 1 */ Set(Value(vals[0]), Register0)
					, /* # 2 */ Set(Memory(0), Value(strs[1]))
					, /* # 3 */ Set(Value(vals[3]), Memory(0))
					, /* # 4 */ Div(Value(strs[0]), Memory(0), Memory(1))
					, /* # 6 */ Set(Value(vals[1]), Register0)
					, /* # 7 */ Set(Value(vals[2]), Memory(1))
					, /* # 8 */ Mul(Register0, Memory(0)	)
					, /* # 9 */ Add(Register0, Memory(1)	)
					, /* #10 */ Set(Value(vals[4]), Register0)
					};
		size_t		memories_n = 2
					, commands_n = (sizeof(commands) / sizeof(commands[0]))
					, result_sz = sizeof(result)
					, int_digits = MAX_Digits
					, fra_digits = MAX_Digits
					, pc = 0
					,x, y, i
					;

		printf("{{{ Test for `Mul_op' and `Div_op' mnemonics starts...\n")
		; (void)fflush(stdout)
		;

		for (x = 0; x < (sizeof(strs) / sizeof(strs[0])); ++x) {
			for (y = 0; y < (sizeof(strs) / sizeof(strs[0])); ++y) {
				/* initialize buffers... */
				memset(result, (int)'@', (sizeof(result) - 1))
				; result[sizeof(result) - 1] = '\0';
				for (i = 0; i < sizeof(vals) / sizeof(vals[0]); ++i) {
					memcpy(vals[i], result, sizeof(vals[0]));
				}

				/* set operands */
				commands[0].operands[0]	= Value(strs[x]);
				commands[0].operands[1]	= Value(strs[y]);
				commands[2].operands[1]	= Value(strs[y]);
				commands[4].operands[0]	= Value(strs[x]);

				/* print operands */
				printf("operands=[ %s, %s ]\n", strs[x], strs[y])
				; (void)fflush(stdout)
				;

				/* don't execute `bc' command, when profiling */
				if (ac != 2 || strcmp(av[1], "-P") != 0) {
					/* printf right answer using `bc(1)' */
					retcode = system(Make_bc_cmd);
					if ((WIFSIGNALED(retcode) != 0)
					&& (WTERMSIG(retcode) == SIGINT
						|| WTERMSIG(retcode) == SIGQUIT
						)
					) {
						char	*signal_notation =
											strsignal(WTERMSIG(retcode))
											;

						printf("!!!system(\"%s\") caught `%s' signal.\n"
								, bc_cmd
								, signal_notation
								); (void)fflush(stdout)
								;

						/*!!!FATAL EXIT!!!*/
						exit(EXIT_FAILURE);
					}
				}

				/* invoke abacus() API */
				errno	= 0;
				retcode	= abacus(result,		result_sz
								, int_digits,	fra_digits
								, memories_n,	commands_n
								, commands,		&pc
								);
				printf(	"abacus() returns %d. (outputs result=[%s](len=%ld)"
						", pc=%ld, errno=%d)\n"
						, retcode
						, result, (long)strlen(result)
						, (long)pc, errno
						);
				if (retcode != NORMAL_END) {
					printf("!!!ERROR!!![%s]\n", strerror(errno))
					; (void)fflush(stdout)
					;
				}

				/* print values in progress */
				for (i = 0; i < sizeof(vals) / sizeof(vals[0]); ++i) {
					printf("%s = vals[%ld]\n", vals[i], (long)i)
					; (void)fflush(stdout)
					;
				}
				(void)puts(""); (void)fflush(stdout);
			}
		}

		printf("}}} Test for `Mul_op' and `Div_op' mnemonics finished.\n")
		; (void)fflush(stdout)
		;
	} /* test `MUL_op' and `Div_op' mnemonics */

	(void)puts(""); (void)fflush(stdout);

	{ /* test `Rou_op' mnemonic */
		char				*nums[] = {
								  "123456789012345678.123456789012345678"
								, "987654321098765432.987654321098765432"
								};
		ssize_t				exp = 0;
		How_to_rounding_e	how = ROUND_up;
		Command_t			commands[] = {
							  /* # 0 */ Rou(nums[0], Value(exp), Value(how))
							};
		size_t				i, pc;

		printf("{{{ Test for `Rou_op' mnemonics finished.\n")
		; (void)fflush(stdout)
		;

		for (exp = (ssize_t)(-MAX_Digits)
		; exp < (ssize_t)(MAX_Digits)
		; ++exp
		) {
		for (how = ROUND_up
		; how <= ROUND_down
		; how = (How_to_rounding_e)(how + 1)
		) {
		for (i = 0; i < sizeof(nums) / sizeof(nums[0]); ++i) {
			/* initialize result buffer */
			memset(result, (int)'@', sizeof(result) - 1);
			result[sizeof(result) - 1]	= '\0';

			/* set up commands[] */
			commands[0].operands[0] = Value(nums[i]);
			commands[0].operands[1] = Value(exp);
			commands[0].operands[2] = Value(how);

			errno	= 0; /* initialize errno */
			pc		= 0; /* initialize program-counter */

			retcode	= abacus(result, sizeof(result)
							, MAX_Digits, MAX_Digits
							, 0
							, (sizeof(commands) / sizeof(commands[0]))
							, commands
							, &pc
							);
			printf("[exp=%ld, how=%s] abacus() returns %d "
				"(errno=%d [%s]).\n"
				"(number=%37.37s)\n"
				"(result=%37.37s)\n"
				, (long)exp
				, ((how == ROUND_up)?		"Round_up"
					: (how == ROUND_even)?	"Round_even"
					:						"Round_down"
					)
				, retcode
				, errno, strerror(errno)
				, nums[i]
				, result
				); (void)fflush(stdout);
		}}}

		printf("}}} Test for `Rou_op' mnemonics finished.\n"
			); (void)fflush(stdout);
	} /* test `Rou_op' mnemonic */
	} /* iterate the unit test for gprof prfiling... */

	exit(EXIT_SUCCESS); /* exit test program for `abacus()' API */
}
