/***********************************************************************
 * File: aggregate.h
 * 	(Extracted from
 *	   "$Id: aggregate.h,v 1.11 2009/06/21 15:37:56 maoyam Exp maoyam $")
 *
 * Abstracts:
 *  This C header file contains definitions of `aggregate()' API.
 *
 * Copyright (C) 2009 Maoyam Tokyo, Japan (mailto:maoyam@mail.goo.ne.jp)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 **********************************************************************/

#ifndef _AGGREGATE_H_ /* { */
#define _AGGREGATE_H_
#ifdef __cplusplus
extern "C" {
#endif

/* abacus needs the bellow system include files */
#include <unistd.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <memory.h>
#include <errno.h>
#include <math.h>

/* maximum length of maximum bit-width integer */
#define MAX_LEN_OF_MAX_BIT_WIDTH_INTEGER	(21)

/* precision of result of `logl(3M)' */
#define PRECISON_OF_LOGL_MATH_LIB			(15)

/* definitions of ope-code for `aggregate_numbers()' API */
typedef enum aggregate_e_opcode {
	  Aggregate_e_opcode_add	/* add the numbers */
	, Aggregate_e_opcode_sub	/* subtract the numbers */
	, Aggregate_e_opcode_mul	/* multiplies the numbers */
	, Aggregate_e_opcode_div	/* divide the numbers */
	, Aggregate_e_opcode_n
} Aggregate_e_opcode;

/* API: aggregate the numeral strings
 * this returns address of the allocated result buffer.
 */
/*@null@*/ extern char *aggregate_numbers(Aggregate_e_opcode	opcode
										, const char			**numerals
										, size_t				n_numerals
										, ssize_t				scale
										);

/* API: release the buffer that is allocated by `aggregate_numbers()' */
extern int destroy_aggregate_numbers(char	*aggregate_numbers_result);

/* API: compare two numeral strings */
/*@null@*/ extern int *compare_numbers(const char		*a_str
										, const char	*b_str
										, int			*result
										);

/* API: predicateion: is the numeral string zero or not? */
extern int is_number_zero(const char	*num);

/* API: duplicate the numeral string (allocate another new string) */
/*@null@*/ extern char *duplicate_numbers(const char	*num);

/* API: round the numeral string */
typedef enum aggregate_e_round_how {
	  Aggregate_e_round_even	= 0	/* rounding even					*/
	, Aggregate_e_round_up			/* rounding up						*/
	, Aggregate_e_round_down		/* rounding down					*/
	, N_Aggregate_e_round_how		/* number of rounding directives	*/
} Aggregate_e_round_how;

/*@null@*/ extern char *round_numbers(Aggregate_e_round_how	how
									, const char			*num
									, ssize_t				round_scale
									);

/* API: absolute the numeral string */
/*@null@*/ extern char *absolute_numbers(const char	*num
										, ssize_t	scale
										);

/* API: make zero */
/*@null@*/ extern char *make_zero_numbers(ssize_t	scale);

/* API: make power of ten */
/*@null@*/ extern char *make_power_of_ten(ssize_t	scale);

/* API: shift the numeral string */
/*@null@*/ extern char *shift_numbers(const char	*num, ssize_t	scale);

/* API: extraction square root	*/
/*@null@*/ extern char *sqrt_numbers(const char	*num, ssize_t	scale);

/* API: exponent of real numbers */
/*@null@*/ extern char	*exponent_numbers(const char	*num
										, ssize_t	scale
										);

/* API: Napierian logarithm for real numbers */
/*@null@*/ extern char *ln_numbers(const char	*num, ssize_t	scale);

#ifdef __cplusplus
}
#endif
#endif /* } !defined(_AGGREGATE_H_) */
